# -*- coding:utf-8 -*-

# Standard libraries.
from cpython cimport Py_buffer
from libc.stdint cimport uint32_t

# Local imports.
from .libpcre2 cimport *
from .pattern cimport Pattern


cdef class Match:
    cdef pcre2_match_data_t *_mtch
    cdef Pattern _pattern
    cdef Py_buffer *_subj
    cdef size_t _ofst # Byte offset, regardless of subject type.
    cdef uint32_t _opts

    @staticmethod
    cdef Match _from_data(
        pcre2_match_data_t *mtch, Pattern pattern, Py_buffer *subj, size_t ofst, uint32_t opts
    )



# -*- coding:utf-8 -*-

# Standard libraries.
from libc.stdint cimport uint32_t
from libc.stdlib cimport malloc, free
from cpython cimport Py_buffer
from cpython cimport array
from cpython.unicode cimport PyUnicode_Check
from cpython.memoryview cimport PyMemoryView_FromMemory

# Local imports.
from .utils cimport *
from .libpcre2 cimport *
from .match cimport Match
from .pattern cimport Pattern
from .consts import BsrChar, NewlineChar


cdef class Scanner:
    """ Iterator object that scans a subject all non-overlapping matches of a
    pattern. Attributes defined in scanner.pxd, see below for an overview:
        _pattern: Pattern object to use for matching
        _subj: Subject to scan
        _is_crlf_newline: Whether the character sequence CRLF denotes a newline
        _is_patn_utf: Whether the pattern was compiled with UTF support
        _state_opts: Options to pass to match
        _state_ofst: Byte offset to match at
        _state_obj_ofst: Object offset to match at
    """


    # =================================== #
    #         Lifetime management         #
    # =================================== #

    def __cinit__(self):
        self._pattern = None
        self._subj = NULL

        self._is_patn_utf = False
        self._is_crlf_newline = False

        self._state_opts = 0
        self._state_ofst = 0
        self._state_obj_ofst = 0


    def __init__(self, *args, **kwargs):
        # Prevent accidental instantiation from normal Python code since we
        # cannot pass pointers into a Python constructor.
        module = self.__class__.__module__
        qualname = self.__class__.__qualname__
        raise TypeError(f"Cannot create '{module}.{qualname}' instances")


    def __dealloc__(self):
        if self._subj is not NULL:
            free_buffer(self._subj)


    @staticmethod
    cdef Scanner _from_data(Pattern pattern, Py_buffer *subj, size_t offset):
        """ Factory function to create Scanner objects from C-type fields. The
        ownership of the given pointers are stolen, which causes the extension
        type to free them when the object is deallocated.
        """
        # Fast call to __new__() that bypasses the __init__() constructor.
        cdef Scanner scanner = Scanner.__new__(Scanner)
        scanner._pattern = pattern
        scanner._subj = subj

        patn_opts = Pattern._info_uint(pattern._code, PCRE2_INFO_ALLOPTIONS)
        scanner._is_patn_utf = (patn_opts & PCRE2_UTF) != 0
        newline = Pattern._info_uint(pattern._code, PCRE2_INFO_NEWLINE)
        scanner._is_crlf_newline = (
            newline == PCRE2_NEWLINE_ANY or
            newline == PCRE2_NEWLINE_CRLF or
            newline == PCRE2_NEWLINE_ANYCRLF
        )
        scanner._state_opts = 0

        # Compute and set byte equivalent offset.
        if scanner._is_patn_utf:
            ofst, obj_ofst = codepoint_to_codeunit(scanner._subj, offset, 0, 0)
            scanner._state_ofst = ofst
            scanner._state_obj_ofst = obj_ofst
        else:
            scanner._state_obj_ofst = offset
            scanner._state_ofst = scanner._state_obj_ofst
        return scanner


    # ======================================== #
    #         Iteration implementation         #
    # ======================================== #

    def __iter__(self):
        return self


    def __next__(self):
        """ Yields next match object found in subject.
        """
        if self._state_obj_ofst > self._subj.len:
            raise StopIteration

        # Attempt match of pattern onto subject.
        match_rc = <int>0
        mtch = Pattern._match(
            self._pattern._code, self._subj, self._state_ofst, self._state_opts, &match_rc
        )

        # Handle no matches in result.
        if match_rc == PCRE2_ERROR_NOMATCH:
            # Default match is not achored so if no match found at current offset, then there
            # will not be any ahead either.
            if self._state_opts == 0:
                pcre2_match_data_free(mtch)
                raise StopIteration

            # Reset options so empty strings can match at next offset.
            self._state_opts = 0

            # Increment to next character and handle possible CRLF newlines.
            obj_ofst_increment = 1
            if self._is_crlf_newline and (self._state_ofst + 1) < self._subj.len:
                if (<bytes>self._subj.buf)[self._state_ofst:self._state_ofst + 2] == b"\r\n":
                    obj_ofst_increment += 1

            # Convert indices accordingly.
            if self._is_patn_utf:
                self._state_ofst, self._state_obj_ofst = codepoint_to_codeunit(
                    self._subj,
                    self._state_obj_ofst + obj_ofst_increment,
                    self._state_ofst,
                    self._state_obj_ofst
                )
            else:
                self._state_obj_ofst = self._state_obj_ofst + obj_ofst_increment
                self._state_ofst = self._state_obj_ofst

            pcre2_match_data_free(mtch)
            return self.__next__()

        # Handle all other errors.
        elif mtch is NULL or match_rc < 0:
            pcre2_match_data_free(mtch)
            raise_from_rc(match_rc, None)

        # If the match was successful.
        else:
            ovec_table = pcre2_get_ovector_pointer(mtch)
            mtch_end = ovec_table[1]

            if self._state_ofst == mtch_end:
                # If the matched string is empty ensure next is not.
                self._state_opts = PCRE2_NOTEMPTY_ATSTART | PCRE2_ANCHORED
            else:
                # Convert the end in the byte string to the end in the object.
                self._state_opts = 0
                if self._is_patn_utf:
                    self._state_ofst, self._state_obj_ofst = codeunit_to_codepoint(
                        self._subj, mtch_end, self._state_ofst, self._state_obj_ofst
                    )
                else:
                    self._state_obj_ofst = mtch_end
                    self._state_ofst = self._state_obj_ofst

            # Create new buffer for match object to own
            subj_copy = get_buffer(self._subj.obj)
            return Match._from_data(
                mtch, self._pattern, subj_copy, self._state_ofst, self._state_opts
            )



# -*- coding:utf-8 -*-

# Standard libraries.
from cpython cimport Py_buffer
from libc.stdint cimport uint32_t

# Local imports.
from .libpcre2 cimport *
from .pattern cimport Pattern


cdef class Scanner:
    cdef Pattern _pattern
    cdef Py_buffer *_subj

    cdef bint _is_crlf_newline
    cdef bint _is_patn_utf

    cdef uint32_t _state_opts
    cdef size_t _state_ofst
    cdef size_t _state_obj_ofst

    @staticmethod
    cdef Scanner _from_data(
        Pattern pattern, Py_buffer *subject, size_t offset
    )



# -*- coding:utf-8 -*-

# Standard libraries.
from cpython cimport Py_buffer


cdef int free_buffer(Py_buffer *pybuf)

cdef Py_buffer * get_buffer(object obj) except NULL

cdef (size_t, size_t) codeunit_to_codepoint(
    Py_buffer *pybuf,
    size_t codeunit_idx,
    size_t cur_codeunit_idx, size_t cur_codepoint_idx
)
cdef (size_t, size_t) codepoint_to_codeunit(
    Py_buffer *pybuf,
    size_t codepoint_idx,
    size_t cur_codeunit_idx, size_t cur_codepoint_idx
)

cdef void * raise_from_rc(int errorcode, object context_msg) except NULL



# -*- coding:utf-8 -*-

# Standard libraries.
from libc.stdlib cimport malloc, free
from libc.stdint cimport uint8_t
from cpython cimport Py_buffer
from cpython.buffer cimport (
    PyObject_CheckBuffer,
    PyBuffer_IsContiguous,
    PyObject_GetBuffer,
    PyBuffer_FillInfo,
    PyBuffer_Release
)
from cpython.unicode cimport (
    PyUnicode_Check
)
cdef extern from "Python.h":
    int PyUnicode_1BYTE_KIND
    int PyUnicode_2BYTE_KIND
    int PyUnicode_4BYTE_KIND
    unsigned int PyUnicode_KIND(object o)
    void *PyUnicode_DATA(object o)
    const char * PyUnicode_AsUTF8AndSize(object unicode, Py_ssize_t *size)

# Local imports.
from .libpcre2 cimport *
from .exceptions import LibraryError, CompileError, MatchError


cdef int free_buffer(Py_buffer *pybuf):
    """ Safe free a buffer pointer, ensuring it first releases it's reference
    """
    if pybuf is not NULL:
        PyBuffer_Release(pybuf)
        free(pybuf)
    return 0


cdef Py_buffer * get_buffer(object obj) except NULL:
    """ Get a Python buffer from an object, encoding via UTF-8 if unicode
    based
    """
    cdef const char *sptr = NULL
    cdef Py_ssize_t length = 0

    pybuf = <Py_buffer *>malloc(sizeof(Py_buffer))
    if not pybuf:
        raise MemoryError()

    # Process unicode and derivative objects.
    if PyUnicode_Check(obj):
        sptr = PyUnicode_AsUTF8AndSize(obj, &length)
        fill_buf_rc = PyBuffer_FillInfo(pybuf, obj, <void *>sptr, length, 1, 0)
        if fill_buf_rc < 0:
            free_buffer(pybuf)
            raise ValueError("Could not fill internal buffer")
    
    # Handle all other bytes-like objects.
    else:
        if PyObject_CheckBuffer(obj):
            get_buffer_rc = PyObject_GetBuffer(obj, pybuf, 0)
            if not PyBuffer_IsContiguous(pybuf, b"A"):
                free_buffer(pybuf)
                raise ValueError("Bytes-like object must be contiguous")
        else:
            free(pybuf)
            raise ValueError("Input must be string or bytes-like")

    return pybuf


cdef (size_t, size_t) codeunit_to_codepoint(
    Py_buffer *pybuf,
    size_t codeunit_idx,
    size_t cur_codeunit_idx, size_t cur_codepoint_idx
):
    """ Convert a code unit index to a code point index
    """
    while cur_codeunit_idx < codeunit_idx:
        if (((<uint8_t *>pybuf.buf)[cur_codeunit_idx]) & 0xC0) != 0x80:
            cur_codepoint_idx += 1
        cur_codeunit_idx += 1
    return cur_codeunit_idx, cur_codepoint_idx

    
cdef (size_t, size_t) codepoint_to_codeunit(
    Py_buffer *pybuf,
    size_t codepoint_idx,
    size_t cur_codeunit_idx, size_t cur_codepoint_idx
):
    """
    """
    while cur_codepoint_idx < codepoint_idx:
        cur_codeunit_idx += 1
        if (((<uint8_t *>pybuf.buf)[cur_codeunit_idx]) & 0xC0) != 0x80:
            cur_codepoint_idx += 1
    return cur_codeunit_idx, cur_codepoint_idx


cdef void * raise_from_rc(int errorcode, object context_msg) except NULL:
    """ Raise the appropriate error type from the given error code

    Raises one of the custom exception classes defined in this module. Each
    exception corresponds to a set of error codes defined in PCRE2. Error
    messages are retrieved from PCRE2.
    """
    # Match against error code classes.
    if errorcode > 0:
        raise CompileError(errorcode, context_msg)

    elif errorcode == PCRE2_ERROR_NOMATCH or errorcode == PCRE2_ERROR_PARTIAL:
        raise MatchError(errorcode, context_msg)

    else:
        raise LibraryError(errorcode, context_msg)



# -*- coding:utf-8 -*-

# Standard libraries.
from enum import IntEnum
from libc.stdint cimport uint32_t
from libc.stdlib cimport malloc, free
from cpython.unicode cimport PyUnicode_Check
cimport cython

# Local imports.
from .utils cimport *
from .libpcre2 cimport *
from .pattern cimport Pattern


@cython.freelist(8)
cdef class Match:
    """
    Object wrapper for a match block in PCRE2. Contains all relevant
    information of a successful match. Attributes defined in match.pxd, see
    below for an overview:
        _mtch: Raw match data block, managed by PCRE2
        _pattern: Pattern object used in match
        _subj: Subject the pattern was matched against
        _ofst: Byte offset (egardless of subject type) used in  match
        _opts: Option bits used in match call
    """

    # =================================== #
    #         Lifetime management         #
    # =================================== #

    def __cinit__(self):
        self._mtch = NULL
        self._pattern = None
        self._subj = NULL
        self._opts = 0


    def __init__(self, *args, **kwargs):
        # Prevent accidental instantiation from normal Python code since we
        # cannot pass pointers into a Python constructor.
        module = self.__class__.__module__
        qualname = self.__class__.__qualname__
        raise TypeError(f"Cannot create '{module}.{qualname}' instances")


    def __dealloc__(self):
        if self._subj is not NULL:
            free_buffer(self._subj)
        if self._mtch is not NULL:
            pcre2_match_data_free(self._mtch)


    @staticmethod
    cdef Match _from_data(
            pcre2_match_data_t *mtch,
            Pattern pattern,
            Py_buffer *subj,
            size_t ofst,
            uint32_t opts):
        """ Factory function to create Match objects from C-type fields. The
        ownership of the given pointers are stolen, which causes the extension
        type to free them when the object is deallocated.
        """

        # Fast call to __new__() that bypasses the __init__() constructor.
        cdef Match match = Match.__new__(Match)
        match._mtch = mtch
        match._pattern = pattern
        match._subj = subj
        match._ofst = ofst # Code unit offset
        match._opts = opts
        return match


    # ========================== #
    #         Properties         #
    # ========================== #

    @property
    def options(self):
        return self._opts

    
    @property
    def subject(self):
        return self._subj.obj

    
    @property
    def pattern(self):
        return self._pattern


    # ======================= #
    #         Methods         #
    # ======================= #

    def start(self, group=0):
        """ Get the starting index of the matched substring, or of a specified
        captured group.
        """
        ovec_count = pcre2_get_ovector_count(self._mtch)
        ovec_table = pcre2_get_ovector_pointer(self._mtch)
        
        cdef int grp_num
        cdef pcre2_sptr_t first_entry
        cdef pcre2_sptr_t last_entry
        if isinstance(group, int):
            grp_num = group
        else:
            grp_name = get_buffer(group)
            pcre2_substring_nametable_scan(
                self._pattern._code, <pcre2_sptr_t>grp_name.buf, &first_entry, &last_entry
            )
            grp_num = (first_entry[0] << 8) | first_entry[1]
            if grp_num < 0:
                raise_from_rc(grp_num, None)
            free_buffer(grp_name)

        if grp_num > <int>ovec_count:
            raise ValueError("Group referenced out of bounds")
        start = ovec_table[2 * grp_num]

        # Convert to code unit index as necessary.
        if PyUnicode_Check(self._subj.obj):
            _, start = codeunit_to_codepoint(self._subj, start, 0, 0)

        return start


    def end(self, group=0):
        """ Get the ending index of the matched substring, or of a specified
        captured group.
        """
        ovec_count = pcre2_get_ovector_count(self._mtch)
        ovec_table = pcre2_get_ovector_pointer(self._mtch)
        
        cdef int grp_num
        cdef pcre2_sptr_t first_entry
        cdef pcre2_sptr_t last_entry
        if isinstance(group, int):
            grp_num = group
        else:
            grp_name = get_buffer(group)
            pcre2_substring_nametable_scan(
                self._pattern._code, <pcre2_sptr_t>grp_name.buf, &first_entry, &last_entry
            )
            grp_num = (first_entry[0] << 8) | first_entry[1]
            if grp_num < 0:
                raise_from_rc(grp_num, None)
            free_buffer(grp_name)

        if grp_num > <int>ovec_count:
            raise ValueError("Group referenced out of bounds.")
        end = ovec_table[2 * grp_num + 1]

        # Convert to code unit index as necessary.
        if PyUnicode_Check(self._subj.obj):
            _, end = codeunit_to_codepoint(self._subj, end, 0, 0)

        return end


    def substring(self, group=0, default=""):
        """ Get the full matched substring, or that of a specified captured
        group.
        """
        cdef uint8_t *res
        cdef size_t res_len
        if isinstance(group, int):
            grp_num = <uint32_t>group

            # Handle unset matches and return default if none found
            is_substring_set = pcre2_substring_length_bynumber(self._mtch, grp_num, NULL)
            if is_substring_set < 0:
                return default

            get_rc = pcre2_substring_get_bynumber(self._mtch, grp_num, &res, &res_len)
            if get_rc < 0:
                raise_from_rc(get_rc, None)
        else:
            grp_name = get_buffer(group)

            # Handle unset matches and return default if none found
            is_substring_set = pcre2_substring_length_byname(
                self._mtch, <pcre2_sptr_t>grp_name.buf, NULL
            )
            if is_substring_set < 0:
                return default

            get_rc = pcre2_substring_get_byname(
                self._mtch, <pcre2_sptr_t>grp_name.buf, &res, &res_len
            )
            if get_rc < 0:
                raise_from_rc(get_rc, None)
            free_buffer(grp_name)

        # Clean up result and convert to unicode as appropriate.
        result = (<pcre2_sptr_t>res)[:res_len]
        result = result.strip(b"\x00")
        if PyUnicode_Check(self._subj.obj):
            result = result.decode("utf-8")
            
        return result


    def __getitem__(self, group):
        """ Alias to substring.
        """
        return self.substring(group)


    def expand(self, replacement, offset=0, options=0, low_memory=False):
        """ Equivalent to calling substitute with the provided match. The type
        of the subject determines the type of the returned string.
        """
        is_subj_utf = <bint>PyUnicode_Check(self._subj.obj)
        is_repl_utf = <bint>PyUnicode_Check(replacement)
        if is_subj_utf ^ is_repl_utf:
            subj_type = "string" if is_subj_utf else "bytes-like"
            repl_type = "string" if is_repl_utf else "bytes-like"
            raise ValueError(f"Cannot use a {subj_type} subject with a {repl_type} replacement")

        # Convert Python objects to C strings.
        repl = get_buffer(replacement)
        cdef size_t obj_ofst = <size_t>offset
        cdef size_t ofst = obj_ofst
        cdef uint32_t opts = <uint32_t>options | PCRE2_SUBSTITUTE_MATCHED
        if is_subj_utf:
            ofst, obj_ofst = codepoint_to_codeunit(self._subj, obj_ofst, 0, 0)

        cdef size_t res_buf_len = 0
        if not low_memory:
            res_buf_len = self._subj.len + (self._subj.len // 2)

        cdef int rc = 0
        res, res_len = Pattern._substitute(
            self._pattern._code, repl, self._subj, res_buf_len, ofst, opts, self._mtch, &rc
        )
        if res is NULL:
            raise_from_rc(rc, None)

        # Clean up result and convert to unicode as appropriate.
        result = (<pcre2_sptr_t>res)[:res_len]
        result = result.strip(b"\x00")
        if is_subj_utf:
            result = result.decode("utf-8")
        
        free(res)
        free_buffer(repl)
        return result

    def groups(self, default=""):
        """ Return a tuple containing all the subgroups of the match, from 1 up to however many
        groups are in the pattern.
        """
        return tuple(self.substring(g, default=default) for g in range(self.pattern.capture_count))



import pytest
import pcre2
from pcre2.exceptions import CompileError, MatchError, LibraryError
from pcre2.consts import CompileOption


test_data_pattern_compile_success = [
    (b"a+b+c*d*", 0, "SUCCESS"),
    (b"(?<foo>a+b+)c*d*", 0, "SUCCESS"),
    (b"(?<foo>a+b+))c*d*", 0, "COMPILE_ERROR"),
    ("å+∫+ç*∂*".encode(), 0, "SUCCESS"),
    ("a+b+c*d*", 0, "SUCCESS"),
    ("(?<foo>a+b+)c*d*", 0, "SUCCESS"),
    ("(?<foo>a+b+))c*d*", 0, "COMPILE_ERROR"),
    ("(?<<foo>a+b+)c*d*", 0, "COMPILE_ERROR"),
    ("(?<foo>a+b+)c*d*(?<foo>a+b+)", 0, "COMPILE_ERROR"),
    ("(?<foo>a+b+)c*d*(?<foo>a+b+)", pcre2.CompileOption.DUPNAMES, "SUCCESS"),
    ("å+∫+ç*∂*", 0, "SUCCESS"),
    ("(?<ƒøø>a+b+)c*d*", 0, "SUCCESS"),
]
@pytest.mark.parametrize("pattern,options,return_code", test_data_pattern_compile_success)
def test_pattern_compile_success(pattern, options, return_code):
    try:
        p = pcre2.compile(pattern, options=options)
        rc = "SUCCESS"
        assert p.jit_size == 0
    except CompileError as e:
        rc = "COMPILE_ERROR"
    except LibraryError as e:
        rc = "LIB_ERROR"
    assert rc == return_code

@pytest.mark.parametrize("pattern,options,return_code", test_data_pattern_compile_success)
def test_pattern_jit_compile_success(pattern, options, return_code):
    try:
        p = pcre2.compile(pattern, options=options, jit=True)
        rc = "SUCCESS"
        assert p.jit_size > 0
    except CompileError as e:
        rc = "COMPILE_ERROR"
    except LibraryError as e:
        rc = "LIB_ERROR"
    assert rc == return_code


test_data_pattern_name_dict = [
    (b"(?<foo>a+b+)c*d*", 0, {1: b"foo"}),
    ("(?<foo>a+b+)c*d*", 0, {1: "foo"}),
    ("(?<ƒøø>a+b+)c*d*", 0, {1: "ƒøø"}),
    ("(?<foo>a+b+)c*d*(?<bar>a+b+)", 0, {1: "foo", 2: "bar"}),
    ("(?<foo>a+b+)c*(.+)d*(?<bar>a+b+)", 0, {1: "foo", 3: "bar"}),
    ("(?<foo>a+b+)c*d*(?<foo>a+b+)", pcre2.CompileOption.DUPNAMES, {1: "foo", 2: "foo"}),
]
@pytest.mark.parametrize("pattern,options,name_dict", test_data_pattern_name_dict)
def test_pattern_name_dict(pattern, options, name_dict):
    p = pcre2.compile(pattern, options=options)
    assert p.name_dict() == name_dict


test_data_pattern_match_success = [
    (b".*", b"abacbaccbacccb", 0, 0, "SUCCESS"),
    (".*", "abacbaccbacccb", 0, 0, "SUCCESS"),
    ("ac{3,}b", "abacbaccbacccb", 0, 0, "SUCCESS"),
    ("a•{3,}b", "aba•ba••ba•••b", 0, 0, "SUCCESS"),
    ("ab", "abacbaccbacccb", 0, 2, "MATCH_ERROR"),
    ("((((((((((((((()))))))))))))))", "", 0, 0, "SUCCESS"),
]
@pytest.mark.parametrize(
    "pattern,subject,options,offset,return_code", test_data_pattern_match_success
)
def test_pattern_match_success(pattern, subject, options, offset, return_code):
    p = pcre2.compile(pattern, options=options)
    try:
        m = p.match(subject, offset=offset)
        rc = "SUCCESS"
    except MatchError as e:
        rc = "MATCH_ERROR"
    except LibraryError as e:
        rc = "LIB_ERROR"
    assert rc == return_code


test_data_pattern_scan_length = [
    (b".+", b"abacbaccbacccb", 0, 1),
    (b".*", b"abacbaccbacccb", 0, 2),
    (".+", "abacbaccbacccb", 0, 1),
    (".*", "abacbaccbacccb", 0, 2),
    ("[abc]*", "dabacbaccbacccb", 0, 3),
    ("ac{2,}b", "abacbaccbacccb", 0, 2),
    ("a•{2,}b", "aba•ba••ba•••b", 0, 2),
    ("a•*b", "aba•ba••ba•••b", 0, 4),
    ("ab", "abacbaccbacccb", 2, 0),
]
@pytest.mark.parametrize(
    "pattern,subject,offset,iter_length", test_data_pattern_scan_length
)
def test_pattern_scan_length(pattern, subject, offset, iter_length):
    p = pcre2.compile(pattern)
    s = p.scan(subject, offset=offset)
    assert len(list(iter(s))) == iter_length


test_pattern_substitute = [
    (b"[abc]*", b"", b"dabacbaccbacccb", False, False, 0, b"dabacbaccbacccb"),
    ("[abc]*", "", "dabacbaccbacccb", False, False, 0, "dabacbaccbacccb"),
    ("[abc]*", "", "dabacbaccbacccb", False, False, 1, "d"),
    ("a(•{2,})b", "a•b", "aba•ba••ba•••b", True, False, 0, "aba•ba•ba•b"),
    ("a(•{2,})b", "a$1b", "aba•ba••ba•••b", True, True, 0, "aba•ba$1ba$1b"),
]
@pytest.mark.parametrize(
    "pattern,replacement,subject,suball,literal,offset,result", test_pattern_substitute
)
def test_pattern_substitute(pattern, replacement, subject, suball, literal, offset, result):
    p = pcre2.compile(pattern)
    assert p.substitute(replacement, subject, suball=suball, literal=literal, offset=offset) == result

def test_pattern_findall():
    p = pcre2.compile(r'(\w+)=(\d+)')
    assert p.findall('set width=20 and height=10') == [('width=20', 'width'), ('height=10', 'height')]
    s = bytes(range(128)).decode()
    p2 = pcre2.compile(r'[0-9--1]')
    assert p2.findall(s) == list('-./0123456789')
    p3 = pcre2.compile(r'[%--1]')
    assert p3.findall(s) == list("%&'()*+,-1")
    p4 = pcre2.compile(r'[%--]')
    assert p4.findall(s) == list("%&'()*+,-")
    p5 = pcre2.compile(r'[0-9&&1]')
    assert p5.findall(s) == list('&0123456789')
    p6 = pcre2.compile(r'[\d&&1]')
    assert p6.findall(s) == list('&0123456789')
    p7 = pcre2.compile(r'[0-9||a]')
    assert p7.findall(s) == list('0123456789a|')
    p8 = pcre2.compile(r'[\d||a]')
    assert p8.findall(s) == list('0123456789a|')
    p9 = pcre2.compile(r'[0-9~~1]')
    assert p9.findall(s) == list('0123456789~')
    p10 = pcre2.compile(r'[\d~~1]')
    assert p10.findall(s) == list('0123456789~')
    p11 = pcre2.compile(r'[[0-9]|]')
    assert p11.findall(s) == list('0123456789[]')

    for reps in '*', '+', '?', '{1}':
        for mod in '', '?':
            pattern = '.' + reps + mod + 'yz'
            assert pcre2.compile(pattern, pcre2.S).findall('xyz') == ['xyz'], pattern
            pattern = pattern.encode()
            assert pcre2.compile(pattern, pcre2.S).findall(b'xyz') == [b'xyz'], pattern


def test_pattern_jit_findall():
    assert pcre2.findall(r'(\w+)=(\d+)', 'set width=20 and height=10') == [('width=20', 'width'), ('height=10', 'height')]
    assert pcre2.findall(":+", "abc") == []
    assert pcre2.findall(":+", "a:b::c:::d") == [":", "::", ":::"]
    assert pcre2.findall("(:+)", "a:b::c:::d") == [":", "::", ":::"]

    for x in ("\xe0", "\u0430", "\U0001d49c"):
        xx = x * 2
        xxx = x * 3
        string = "a%sb%sc%sd" % (x, xx, xxx)
        assert pcre2.findall("%s+" % x, string) == [x, xx, xxx]
        assert pcre2.findall("(%s+)" % x, string) == [x, xx, xxx]

    assert len(pcre2.findall(r"\b", "a")) == 2
    assert len(pcre2.findall(r"\B", "a")) == 0
    assert len(pcre2.findall(r"\b", " ")) == 0
    assert len(pcre2.findall(r"\b", "   ")) == 0
    assert len(pcre2.findall(r"\B", " ")) == 2

    s = bytes(range(128)).decode()
    assert pcre2.findall(r'[--1]', s) ==  list('-./01')
    assert pcre2.findall(r'[&&1]', s) ==  list('&1')
    assert pcre2.findall(r'[||1]', s) ==  list('1|')
    assert pcre2.findall(r'[~~1]', s) ==  list('1~')

    assert pcre2.findall(r"(?i)(a)\1", "aa \u0100") == ['a']

    assert pcre2.findall(r'a++', 'aab') == ['aa']
    assert pcre2.findall(r'a*+', 'aab') == ['aa', '', '']
    assert pcre2.findall(r'a?+', 'aab') == ['a', 'a', '', '']
    assert pcre2.findall(r'a{1,3}+', 'aab') == ['aa']

    assert pcre2.findall(r'(?:ab)++', 'ababc') == ['abab']
    assert pcre2.findall(r'(?:ab)*+', 'ababc') == ['abab', '', '']
    assert pcre2.findall(r'(?:ab)?+', 'ababc') == ['ab', 'ab', '', '']
    assert pcre2.findall(r'(?:ab){1,3}+', 'ababc') == ['abab']

    assert pcre2.findall(r'(?>a+)', 'aab') == ['aa']
    assert pcre2.findall(r'(?>a*)', 'aab') == ['aa', '', '']
    assert pcre2.findall(r'(?>a?)', 'aab') == ['a', 'a', '', '']
    assert pcre2.findall(r'(?>a{1,3})', 'aab') == ['aa']

    assert pcre2.findall(r'(?>(?:ab)+)', 'ababc') == ['abab']
    assert pcre2.findall(r'(?>(?:ab)*)', 'ababc') == ['abab', '', '']
    assert pcre2.findall(r'(?>(?:ab)?)', 'ababc') == ['ab', 'ab', '', '']
    assert pcre2.findall(r'(?>(?:ab){1,3})', 'ababc') == ['abab']

    import re
    b = 'y\u2620y\u2620y'.encode('utf-8')
    assert len(pcre2.findall(re.escape('\u2620'.encode('utf-8')), b)) == 2


def test_pattern_split():
    pattern = "[\u002E\u3002\uFF0E\uFF61]"
    assert pcre2.compile(pattern).split("a.b.c") == ['a','b','c']


def test_pattern_jit_split():
    assert pcre2.split(":", ":a:b::c") == ['', 'a', 'b', '', 'c']
    assert pcre2.split(":+", ":a:b::c") == ['', 'a', 'b', 'c']
    assert pcre2.split("(:+)", ":a:b::c") == ['', ':', 'a', ':', 'b', '::', 'c']

    assert pcre2.split(b":", b":a:b::c") == [b'', b'a', b'b', b'', b'c']
    assert pcre2.split(b":+", b":a:b::c") == [b'', b'a', b'b', b'c']
    assert pcre2.split(b"(:+)", b":a:b::c") == [b'', b':', b'a', b':', b'b', b'::', b'c']

    for a, b, c in ("\xe0\xdf\xe7", "\u0430\u0431\u0432",
                    "\U0001d49c\U0001d49e\U0001d4b5"):
        string = ":%s:%s::%s" % (a, b, c)
        assert pcre2.split(":", string) == ['', a, b, '', c]
        assert pcre2.split(":+", string) == ['', a, b, c]
        assert pcre2.split("(:+)", string) == ['', ':', a, ':', b, '::', c]

    assert pcre2.split("(?::+)", ":a:b::c") == ['', 'a', 'b', 'c']
    assert pcre2.split("([b:]+)", ":a:b::c") == ['', ':', 'a', ':b::', 'c']
    assert pcre2.split("(?:b)|(?::+)", ":a:b::c") == ['', 'a', '', '', 'c']

    assert pcre2.split(":", ":a:b::c", 2) == ['', 'a', 'b::c']
    assert pcre2.split(":", ":a:b::c", maxsplit=2) == ['', 'a', 'b::c']
    assert pcre2.split(':', 'a:b:c:d', maxsplit=2) == ['a', 'b', 'c:d']
    assert pcre2.split("(:)", ":a:b::c", maxsplit=2) == ['', ':', 'a', ':', 'b::c']
    assert pcre2.split("(:+)", ":a:b::c", maxsplit=2) == ['', ':', 'a', ':', 'b::c']



# -*- coding:utf-8 -*-

# Standard libraries.
from libc.stdint cimport uint32_t
from cpython cimport Py_buffer
from cpython.unicode cimport PyUnicode_Check

# Local imports.
from .utils cimport *
from .libpcre2 cimport *
from .pattern cimport Pattern
from .match cimport Match


def compile(pattern, options=0, jit=False):
    """ Factory function to compile regular expressions into Pattern objects.
    See the following PCRE2 documentation for a brief overview of the relevant
    options:
        http://pcre.org/current/doc/html/pcre2_compile.html
    """
    
    cdef Py_buffer *patn = get_buffer(pattern)
    cdef uint32_t opts = <uint32_t>options

    # Ensure unicode strings are processed with UTF-8 support.
    if PyUnicode_Check(pattern):
        opts = opts | PCRE2_UTF

    cdef int compile_rc
    cdef size_t compile_errpos
    cdef pcre2_code_t *code = pcre2_compile(
        <pcre2_sptr_t>patn.buf, <size_t>patn.len, opts, &compile_rc, &compile_errpos, NULL
    )

    if code is NULL:
        # If source was a unicode string, use the code point offset.
        if PyUnicode_Check(pattern):
            _, compile_errpos = codeunit_to_codepoint(patn, compile_errpos, 0, 0)
        additional_msg = f"Compilation failed at position {compile_errpos!r}"
        raise_from_rc(compile_rc, additional_msg)

    pattern_obj = Pattern._from_data(code, patn, opts)
    if jit:
        pattern_obj.jit_compile()
    return pattern_obj


def findall(pattern, subject, offset=0, options=0, jit=True):
    """ Shorthand for compiling a pattern, then calling findall. Note that this
    will use JIT compilation.
    """
    return compile(pattern, options=options, jit=jit).findall(subject, offset=offset)


def match(pattern, subject, offset=0, options=0, jit=False):
    """ Shorthand for compiling a pattern, then calling match.
    """
    return compile(pattern, options=options, jit=jit).match(subject, offset=offset)


def scan(pattern, subject, offset=0, options=0, jit=True):
    """ Shorthand for compiling a pattern, then calling scan. Note that this
    will use JIT compilation.
    """
    return compile(pattern, options=options, jit=jit).scan(subject, offset=offset)


def split(pattern, subject, maxsplit=0, offset=0, options=0, jit=True):
    """ Shorthand for compiling a pattern, then calling split. Note that this
    will use JIT compilation.
    """
    pattern_obj = compile(pattern, options=options, jit=jit)
    return pattern_obj.split(subject, maxsplit=maxsplit, offset=offset)


def substitute(
    pattern,
    replacement,
    subject,
    offset=0,
    suball=True,
    literal=False,
    low_memory=False,
    options=0,
    jit=True
):
    """ Shorthand for compiling a pattern, then calling substitute.
    """
    pattern_obj = compile(pattern, options=options, jit=jit)
    if suball:
        pattern_obj.jit_compile()
    return pattern_obj.substitute(
        replacement, subject, offset=offset, suball=suball, literal=literal, low_memory=low_memory
    )



import pytest
import pcre2
from pcre2.exceptions import CompileError, MatchError, LibraryError

def test_match_groups():
    assert pcre2.match('a', 'a').groups() == ()
    assert pcre2.match('(a)', 'a').groups() == ('a',)

    assert pcre2.match(b'a', b'a').groups() == ()
    assert pcre2.match(b'(a)', b'a').groups() == (b'a',)

    for a in ("\xe0", "\u0430", "\U0001d49c"):
        assert pcre2.match(a, a).groups() == ()
        assert pcre2.match('(%s)' % a, a).groups() == (a,)



# -*- coding:utf-8 -*-

import os
import skbuild
import setuptools


def get_long_desciption():
    cwd = os.path.abspath(os.path.dirname(__file__))
    filename = os.path.join(cwd, "README.md")
    with open(filename) as f:
        long_description = f.read()

    return long_description


skbuild.setup(
    name = "pcre2",
    version = "0.4.0",
    description = "Python bindings for the PCRE2 regular expression library",
    long_description = get_long_desciption(),
    long_description_content_type = "text/markdown",
    license = "BSD 3-Clause License",
    author = "Garrett Tetrault",
    url = "https://github.com/grtetrault/pcre2.py",
    classifiers = [
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Programming Language :: C",
        "Programming Language :: Cython",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: POSIX :: Linux",
        "Operating System :: Microsoft :: Windows"
    ],
    include_package_data=True,
    packages = setuptools.find_packages("src"),
    package_dir = {"": "src"},
    cmake_languages = "C",
)
